#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
GUI configuration for the base BSM2 flowsheet.
"""

from pyomo.environ import units as pyunits, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent

from watertap.ui.fsapi import FlowsheetInterface

from watertap.core.util.initialization import assert_degrees_of_freedom

from watertap.examples.flowsheets.case_studies.full_water_resource_recovery_facility.BSM2 import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
)


def export_to_ui():
    """
    Exports the variables, flowsheet build, and solver results to the GUI.
    """
    return FlowsheetInterface(
        name="BSM2",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    """
    Exports the variables to the GUI.
    """
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    exports.add(
        obj=fs.FeedWater.flow_vol[0],
        name="Feed volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Inlet volumetric flow rate",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "S_I"],
        name="Feed S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet soluble inert organic matter concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "S_S"],
        name="Feed S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet readily biodegradable substrate concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "X_I"],
        name="Feed X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet particulate inert organic matter concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "X_S"],
        name="Feed X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet slowly biodegradable substrate concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "X_BH"],
        name="Feed X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet active heterotrophic biomass concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "X_BA"],
        name="Feed X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet active autotrophic biomass concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "X_P"],
        name="Feed X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet particulate products arising from biomass decay concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "S_O"],
        name="Feed S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet oxygen concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "S_NO"],
        name="Feed S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet nitrate and nitrite nitrogen concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "S_NH"],
        name="Feed S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet ammonium and ammonia nitrogen concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "S_ND"],
        name="Feed S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet biodegradable organic nitrogen concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.conc_mass_comp[0, "X_ND"],
        name="Feed X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=2,
        description="Inlet particulate biodegradable organic nitrogen concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.FeedWater.properties[0].alkalinity,
        name="Feed S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=2,
        description="Alkalinity",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )

    # Unit model data, activated sludge process
    exports.add(
        obj=fs.R1.volume[0],
        name="First anoxic reactor volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="CSTR volume",
        is_input=True,
        input_category="Activated sludge process",
        is_output=False,
    )
    exports.add(
        obj=fs.R2.volume[0],
        name="Second anoxic reactor volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="CSTR volume",
        is_input=True,
        input_category="Activated sludge process",
        is_output=False,
    )
    exports.add(
        obj=fs.R3.volume[0],
        name="First aerobic reactor volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="CSTR volume",
        is_input=True,
        input_category="Activated sludge process",
        is_output=False,
    )
    exports.add(
        obj=fs.R4.volume[0],
        name="Second aerobic reactor volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="CSTR volume",
        is_input=True,
        input_category="Activated sludge process",
        is_output=False,
    )
    exports.add(
        obj=fs.R5.volume[0],
        name="Third aerobic reactor volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="CSTR volume",
        is_input=True,
        input_category="Activated sludge process",
        is_output=False,
    )

    # Unit model data, anaerobic digestor
    exports.add(
        obj=fs.RADM.volume_liquid[0],
        name="Liquid volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="Liquid volume",
        is_input=True,
        input_category="Anaerobic digestor",
        is_output=False,
    )
    exports.add(
        obj=fs.RADM.volume_vapor[0],
        name="Vapor volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=1,
        description="Vapor volume",
        is_input=True,
        input_category="Anaerobic digestor",
        is_output=False,
    )

    # Unit model data, secondary clarifier
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "H2O"],
        name="H2O split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Water split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_I"],
        name="S_I split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Soluble inert organic matter split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_S"],
        name="S_S split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Readily biodegradable substrate split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "X_I"],
        name="X_I split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Particulate inert organic matter split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "X_S"],
        name="X_S split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Slowly biodegradable substrate split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "X_BH"],
        name="X_BH split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Active heterotrophic biomass split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "X_BA"],
        name="X_BA split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Active autotrophic biomass split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "X_P"],
        name="X_P split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Particulate products arising from biomass decay split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_O"],
        name="S_O split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Oxygen split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_NO"],
        name="S_NO split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Nitrate and nitrite nitrogen split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_NH"],
        name="S_NH split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Ammonium and ammonia nitrogen split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_ND"],
        name="S_ND split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Soluble biodegradable organic nitrogen split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "X_ND"],
        name="X_ND split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Particulate biodegradable organic nitrogen split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL1.split_fraction[0, "effluent", "S_ALK"],
        name="S_ALK split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Alkalinity split fraction",
        is_input=True,
        input_category="Secondary clarifier",
        is_output=False,
    )
    # Unit model data, primary clarifier
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "H2O"],
        name="H2O split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Water split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_I"],
        name="S_I split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Soluble inert organic matter split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_S"],
        name="S_S split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Readily biodegradable substrate split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "X_I"],
        name="X_I split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Particulate inert organic matter split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "X_S"],
        name="X_S split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Slowly biodegradable substrate split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "X_BH"],
        name="X_BH split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Active heterotrophic biomass split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "X_BA"],
        name="X_BA split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Active autotrophic biomass split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "X_P"],
        name="X_P split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Particulate products arising from biomass decay split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_O"],
        name="S_O split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Oxygen split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_NO"],
        name="S_NO split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Nitrate and nitrite nitrogen split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_NH"],
        name="S_NH split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Ammonium and ammonia nitrogen split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_ND"],
        name="S_ND split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Soluble biodegradable organic nitrogen split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "X_ND"],
        name="X_ND split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Particulate biodegradable organic nitrogen split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )
    exports.add(
        obj=fs.CL.split_fraction[0, "effluent", "S_ALK"],
        name="S_ALK split fraction",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=5,
        description="Alkalinity split fraction",
        is_input=True,
        input_category="Primary clarifier",
        is_output=False,
    )

    # TODO: uncomment and revise below once costing is merged
    # System costing
    # exports.add(
    #     obj=fs.costing.utilization_factor,
    #     name="Utilization factor",
    #     ui_units=pyunits.dimensionless,
    #     display_units="fraction",
    #     rounding=2,
    #     description="Utilization factor - [annual use hours/total hours in year]",
    #     is_input=True,
    #     input_category="System costing",
    #     is_output=False,
    # )
    # exports.add(
    #     obj=fs.costing.TIC,
    #     name="Practical investment factor",
    #     ui_units=pyunits.dimensionless,
    #     display_units="fraction",
    #     rounding=1,
    #     description="Practical investment factor - [total investment cost/direct "
    #     "capital costs]",
    #     is_input=True,
    #     input_category="System costing",
    #     is_output=False,
    # )
    # exports.add(
    #     obj=fs.costing.plant_lifetime,
    #     name="Plant lifetime",
    #     ui_units=pyunits.year,
    #     display_units="years",
    #     rounding=1,
    #     description="Plant lifetime",
    #     is_input=True,
    #     input_category="System costing",
    #     is_output=False,
    # )
    # exports.add(
    #     obj=fs.costing.wacc,
    #     name="Discount rate",
    #     ui_units=pyunits.dimensionless,
    #     display_units="fraction",
    #     rounding=2,
    #     description="Discount rate used in calculating the capital annualization",
    #     is_input=True,
    #     input_category="System costing",
    #     is_output=False,
    # )
    # exports.add(
    #     obj=fs.costing.maintenance_costs_percent_FCI,
    #     name="Fixed operating cost factor",
    #     ui_units=1 / pyunits.year,
    #     display_units="fraction/year",
    #     rounding=2,
    #     description="Fixed operating cost factor - [annual fixed operating cost/total "
    #     "investment cost]",
    #     is_input=True,
    #     input_category="System costing",
    #     is_output=False,
    # )
    # exports.add(
    #     obj=fs.costing.electricity_cost,
    #     name="Electricity cost",
    #     ui_units=fs.costing.base_currency / pyunits.kWh,
    #     display_units="$/kWh",
    #     rounding=3,
    #     description="Electricity cost",
    #     is_input=True,
    #     input_category="System costing",
    #     is_output=False,
    # )

    # Outlets
    exports.add(
        obj=fs.Treated.properties[0].flow_vol,
        name="Treated flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Outlet treated stream flow rate",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["S_I"],
        name="Treated S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["S_S"],
        name="Treated S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["X_I"],
        name="Treated X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["X_S"],
        name="Treated X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["X_BH"],
        name="Treated X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["X_BA"],
        name="Treated X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["X_P"],
        name="Treated X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["S_O"],
        name="Treated S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["S_NO"],
        name="Treated S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["S_NH"],
        name="Treated S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["S_ND"],
        name="Treated S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].conc_mass_comp["X_ND"],
        name="Treated X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )
    exports.add(
        obj=fs.Treated.properties[0].alkalinity,
        name="Treated S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Outlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Effluent",
    )

    exports.add(
        obj=fs.CL.effluent.flow_vol[0],
        name="Primary clarifier effluent flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Outlet primary clarifier flow rate",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "S_I"],
        name="Primary clarifier effluent S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "S_S"],
        name="Primary clarifier effluent S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "X_I"],
        name="Primary clarifier effluent X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "X_S"],
        name="Primary clarifier effluent X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "X_BH"],
        name="Primary clarifier effluent X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "X_BA"],
        name="Primary clarifier effluent X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "X_P"],
        name="Primary clarifier effluent X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "S_O"],
        name="Primary clarifier effluent S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "S_NO"],
        name="Primary clarifier effluent S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "S_NH"],
        name="Primary clarifier effluent S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "S_ND"],
        name="Primary clarifier effluent S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.conc_mass_comp[0, "X_ND"],
        name="Primary clarifier effluent X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )
    exports.add(
        obj=fs.CL.effluent.alkalinity[0],
        name="Primary clarifier effluent S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Outlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Effluent",
    )

    exports.add(
        obj=fs.CL.underflow.flow_vol[0],
        name="Primary clarifier underflow flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Outlet primary clarifier flow rate",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "S_I"],
        name="Primary clarifier underflow S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "S_S"],
        name="Primary clarifier underflow S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "X_I"],
        name="8Primary clarifier underflow X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "X_S"],
        name="Primary clarifier underflow X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "X_BH"],
        name="Primary clarifier underflow X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "X_BA"],
        name="Primary clarifier underflow X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "X_P"],
        name="Primary clarifier underflow X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "S_O"],
        name="Primary clarifier underflow S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "S_NO"],
        name="Primary clarifier underflow S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "S_NH"],
        name="Primary clarifier underflow S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "S_ND"],
        name="Primary clarifier underflow S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.conc_mass_comp[0, "X_ND"],
        name="Primary clarifier underflow X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )
    exports.add(
        obj=fs.CL.underflow.alkalinity[0],
        name="Primary clarifier underflow S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Outlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Primary Clarifier Underflow",
    )

    exports.add(
        obj=fs.SP6.recycle.flow_vol[0],
        name="ASP recycle inlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="ASP recycle flow rate",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "S_I"],
        name="ASP recycle inlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "S_S"],
        name="ASP recycle inlet S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "X_I"],
        name="ASP recycle inlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "X_S"],
        name="ASP recycle inlet X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "X_BH"],
        name="ASP recycle inlet X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "X_BA"],
        name="ASP recycle inlet X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "X_P"],
        name="ASP recycle inlet X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "S_O"],
        name="ASP recycle inlet S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "S_NO"],
        name="ASP recycle inlet S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "S_NH"],
        name="ASP recycle inlet S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "S_ND"],
        name="ASP recycle inlet S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.conc_mass_comp[0, "X_ND"],
        name="ASP recycle inlet X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )
    exports.add(
        obj=fs.SP6.recycle.alkalinity[0],
        name="ASP recycle inlet S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Inlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="ASP Recycle Inlet",
    )

    exports.add(
        obj=fs.SP5.overflow.flow_vol[0],
        name="Secondary clarifier inlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Secondary clarifier inlet flow rate",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "S_I"],
        name="Secondary clarifier inlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "S_S"],
        name="Secondary clarifier inlet S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "X_I"],
        name="Secondary clarifier inlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "X_S"],
        name="Secondary clarifier inlet X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "X_BH"],
        name="Secondary clarifier inlet X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "X_BA"],
        name="Secondary clarifier inlet X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "X_P"],
        name="Secondary clarifier inlet X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "S_O"],
        name="Secondary clarifier inlet S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "S_NO"],
        name="Secondary clarifier inlet S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "S_NH"],
        name="Secondary clarifier inlet S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "S_ND"],
        name="Secondary clarifier inlet S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.conc_mass_comp[0, "X_ND"],
        name="Secondary clarifier inlet X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )
    exports.add(
        obj=fs.SP5.overflow.alkalinity[0],
        name="Secondary clarifier inlet S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Inlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Secondary Clarifier Inlet",
    )

    exports.add(
        obj=fs.TU.inlet.flow_vol[0],
        name="Thickener inlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Thickener inlet flow rate",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "S_I"],
        name="Thickener inlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "S_S"],
        name="Thickener inlet S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "X_I"],
        name="Thickener inlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "X_S"],
        name="Thickener inlet X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "X_BH"],
        name="Thickener inlet X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "X_BA"],
        name="Thickener inlet X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "X_P"],
        name="Thickener inlet X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "S_O"],
        name="Thickener inlet S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "S_NO"],
        name="Thickener inlet S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "S_NH"],
        name="Thickener inlet S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "S_ND"],
        name="Thickener inlet S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.conc_mass_comp[0, "X_ND"],
        name="Thickener inlet X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )
    exports.add(
        obj=fs.TU.inlet.alkalinity[0],
        name="Thickener inlet S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Inlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Thickener Inlet",
    )

    exports.add(
        obj=fs.asm_adm.inlet.flow_vol[0],
        name="ASM1/ADM1 inlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="ASM1/ADM1 interface inlet flow rate",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "S_I"],
        name="ASM1/ADM1 inlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "S_S"],
        name="ASM1/ADM1 inlet S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "X_I"],
        name="ASM1/ADM1 inlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "X_S"],
        name="ASM1/ADM1 inlet X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "X_BH"],
        name="ASM1/ADM1 inlet X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "X_BA"],
        name="ASM1/ADM1 inlet X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "X_P"],
        name="ASM1/ADM1 inlet X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "S_O"],
        name="ASM1/ADM1 inlet S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "S_NO"],
        name="ASM1/ADM1 inlet S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "S_NH"],
        name="ASM1/ADM1 inlet S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "S_ND"],
        name="ASM1/ADM1 inlet S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.conc_mass_comp[0, "X_ND"],
        name="ASM1/ADM1 inlet X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )
    exports.add(
        obj=fs.asm_adm.inlet.alkalinity[0],
        name="ASM1/ADM1 inlet S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Inlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="ASM1/ADM1 Interface Inlet",
    )

    exports.add(
        obj=fs.asm_adm.outlet.flow_vol[0],
        name="Anaerobic digestor inlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Anaerobic digestor inlet flow rate",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_su"],
        name="Anaerobic digestor inlet S_su concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Monosaccharides concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_aa"],
        name="Anaerobic digestor inlet S_aa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Amino acids concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_fa"],
        name="Anaerobic digestor inlet S_fa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Long chain fatty acids concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_va"],
        name="Anaerobic digestor inlet S_va concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total valerate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_bu"],
        name="Anaerobic digestor inlet S_bu concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total butyrate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_pro"],
        name="Anaerobic digestor inlet S_pro concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total propionate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_ac"],
        name="Anaerobic digestor inlet S_ac concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total acetate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_h2"],
        name="Anaerobic digestor inlet S_h2 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Hydrogen gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_ch4"],
        name="Anaerobic digestor inlet S_ch4 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Methane gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_IC"],
        name="Anaerobic digestor inlet S_IC concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inorganic carbon concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_IN"],
        name="Anaerobic digestor inlet S_IN concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inorganic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "S_I"],
        name="Anaerobic digestor inlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Soluble inerts concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_c"],
        name="Anaerobic digestor inlet X_c concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Composites concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_ch"],
        name="Anaerobic digestor inlet X_ch concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Carbohydrates concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_pr"],
        name="Anaerobic digestor inlet X_pr concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Proteins concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_li"],
        name="Anaerobic digestor inlet X_li concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Lipids concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_su"],
        name="Anaerobic digestor inlet X_su concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Sugar degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_aa"],
        name="Anaerobic digestor inlet X_aa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Amino acid degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_fa"],
        name="Anaerobic digestor inlet X_fa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Long chain fatty acid (LCFA) degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_c4"],
        name="Anaerobic digestor inlet X_c4 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Valerate and butyrate degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_pro"],
        name="Anaerobic digestor inlet X_pro concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Propionate degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_ac"],
        name="Anaerobic digestor inlet X_ac concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Acetate degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_h2"],
        name="Anaerobic digestor inlet X_h2 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Hydrogen degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )
    exports.add(
        obj=fs.asm_adm.outlet.conc_mass_comp[0, "X_I"],
        name="Anaerobic digestor inlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Particulate inerts concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Inlet (post-interface)",
    )

    exports.add(
        obj=fs.adm_asm.inlet.flow_vol[0],
        name="Anaerobic digestor liquid outlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Anaerobic digestor liquid outlet flow rate",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_su"],
        name="Anaerobic digestor liquid outlet S_su concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Monosaccharides concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_aa"],
        name="Anaerobic digestor liquid outlet S_aa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Amino acids concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_fa"],
        name="Anaerobic digestor liquid outlet S_fa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Long chain fatty acids concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_va"],
        name="Anaerobic digestor liquid outlet S_va concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total valerate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_bu"],
        name="Anaerobic digestor liquid outlet S_bu concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total butyrate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_pro"],
        name="Anaerobic digestor liquid outlet S_pro concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total propionate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_ac"],
        name="Anaerobic digestor liquid outlet S_ac concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Total acetate concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_h2"],
        name="Anaerobic digestor liquid outlet S_h2 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Hydrogen gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_ch4"],
        name="Anaerobic digestor liquid outlet S_ch4 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Methane gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_IC"],
        name="Anaerobic digestor liquid outlet S_IC concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inorganic carbon concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_IN"],
        name="Anaerobic digestor liquid outlet S_IN concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Inorganic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "S_I"],
        name="Anaerobic digestor liquid outlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Soluble inerts concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_c"],
        name="Anaerobic digestor liquid outlet X_c concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Composites concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_ch"],
        name="Anaerobic digestor liquid outlet X_ch concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Carbohydrates concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_pr"],
        name="Anaerobic digestor liquid outlet X_pr concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Proteins concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_li"],
        name="Anaerobic digestor liquid outlet X_li concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Lipids concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_su"],
        name="Anaerobic digestor liquid outlet X_su concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Sugar degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_aa"],
        name="Anaerobic digestor liquid outlet X_aa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Amino acid degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_fa"],
        name="Anaerobic digestor liquid outlet X_fa concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Long chain fatty acid (LCFA) degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_c4"],
        name="Anaerobic digestor liquid outlet X_c4 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Valerate and butyrate degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_pro"],
        name="Anaerobic digestor liquid outlet X_pro concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Propionate degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_ac"],
        name="Anaerobic digestor liquid outlet X_ac concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Acetate degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_h2"],
        name="Anaerobic digestor liquid outlet X_h2 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Hydrogen degraders concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )
    exports.add(
        obj=fs.adm_asm.inlet.conc_mass_comp[0, "X_I"],
        name="Anaerobic digestor liquid outlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Particulate inerts concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Liquid Outlet (pre-interface)",
    )

    exports.add(
        obj=fs.adm_asm.outlet.flow_vol[0],
        name="ADM1/ASM1 outlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="ADM1/ASM1 interface outlet flow rate",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "S_I"],
        name="ADM1/ASM1 outlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "S_S"],
        name="ADM1/ASM1 outlet S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "X_I"],
        name="ADM1/ASM1 outlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "X_S"],
        name="ADM1/ASM1 outlet X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "X_BH"],
        name="ADM1/ASM1 outlet X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "X_BA"],
        name="ADM1/ASM1 outlet X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "X_P"],
        name="ADM1/ASM1 outlet X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "S_O"],
        name="ADM1/ASM1 outlet S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "S_NO"],
        name="ADM1/ASM1 outlet S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "S_NH"],
        name="ADM1/ASM1 outlet S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "S_ND"],
        name="ADM1/ASM1 outlet S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.conc_mass_comp[0, "X_ND"],
        name="ADM1/ASM1 outlet X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )
    exports.add(
        obj=fs.adm_asm.outlet.alkalinity[0],
        name="ADM1/ASM1 outlet S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Outlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="ADM1/ASM1 Interface Outlet",
    )

    exports.add(
        obj=fs.RADM.vapor_outlet.flow_vol[0],
        name="Anaerobic digestor vapor flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Anaerobic digestor vapor outlet flow rate",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Vapor Outlet",
    )
    exports.add(
        obj=fs.RADM.vapor_outlet.conc_mass_comp[0, "S_h2"],
        name="Anaerobic digestor vapor S_h2 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Hydrogen gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Vapor Outlet",
    )
    exports.add(
        obj=fs.RADM.vapor_outlet.conc_mass_comp[0, "S_ch4"],
        name="Anaerobic digestor vapor S_ch4 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Methane gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Vapor Outlet",
    )
    exports.add(
        obj=fs.RADM.vapor_outlet.conc_mass_comp[0, "S_co2"],
        name="Anaerobic digestor vapor S_co2 concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Carbon dioxide gas concentration",
        is_input=False,
        is_output=True,
        output_category="Anaerobic Digestor Vapor Outlet",
    )

    exports.add(
        obj=fs.DU.underflow.flow_vol[0],
        name="Dewatered sludge flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Outlet sludge disposal flow rate",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "S_I"],
        name="Dewatered sludge S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "S_S"],
        name="Dewatered sludge S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "X_I"],
        name="Dewatered sludge X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "X_S"],
        name="Dewatered sludge X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "X_BH"],
        name="Dewatered sludge X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "X_BA"],
        name="Dewatered sludge X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "X_P"],
        name="Dewatered sludge X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "S_O"],
        name="Dewatered sludge S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "S_NO"],
        name="Dewatered sludge S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "S_NH"],
        name="Dewatered sludge S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "S_ND"],
        name="Dewatered sludge S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.conc_mass_comp[0, "X_ND"],
        name="Dewatered sludge X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )
    exports.add(
        obj=fs.DU.underflow.alkalinity[0],
        name="Dewatered sludge S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Outlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatered Sludge",
    )

    exports.add(
        obj=fs.DU.overflow.flow_vol[0],
        name="Dewatering unit liquid outlet flow rate",
        ui_units=pyunits.m**3 / pyunits.day,
        display_units="m3/day",
        rounding=2,
        description="Dewatering unit overflow flow rate",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "S_I"],
        name="Dewatering unit liquid outlet S_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "S_S"],
        name="Dewatering unit liquid outlet S_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet readily biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "X_I"],
        name="Dewatering unit liquid outlet X_I concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate inert organic matter concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "X_S"],
        name="Dewatering unit liquid outlet X_S concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet slowly biodegradable substrate concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "X_BH"],
        name="Dewatering unit liquid outlet X_BH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active heterotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "X_BA"],
        name="Dewatering unit liquid outlet X_BA concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet active autotrophic biomass concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "X_P"],
        name="Dewatering unit liquid outlet X_P concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate products arising from biomass decay concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "S_O"],
        name="Dewatering unit liquid outlet S_O concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet oxygen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "S_NO"],
        name="Dewatering unit liquid outlet S_NO concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet nitrate and nitrite nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "S_NH"],
        name="Dewatering unit liquid outlet S_NH concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet ammonium and ammonia nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "S_ND"],
        name="Dewatering unit liquid outlet S_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet soluble biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.conc_mass_comp[0, "X_ND"],
        name="Dewatering unit liquid outlet X_ND concentration",
        ui_units=pyunits.g / pyunits.m**3,
        display_units="g/m3",
        rounding=5,
        description="Outlet particulate biodegradable organic nitrogen concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )
    exports.add(
        obj=fs.DU.overflow.alkalinity[0],
        name="Dewatering unit liquid outlet S_ALK concentration",
        ui_units=pyunits.mol / pyunits.m**3,
        display_units="mol/m3",
        rounding=5,
        description="Outlet alkalinity concentration",
        is_input=False,
        is_output=True,
        output_category="Dewatering Unit Liquid Outlet",
    )

    # performance metrics
    recovery_vol = (
        fs.Treated.properties[0].flow_vol / fs.FeedWater.properties[0].flow_vol
    )
    exports.add(
        obj=recovery_vol,
        name="Volumetric recovery",
        ui_units=pyunits.dimensionless,
        display_units="m3 of product/m3 of feed",
        rounding=5,
        description="Normalized volumetric recovery",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )


def build_flowsheet(build_options=None, **kwargs):
    """
    Builds the initial flowsheet.
    """
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)
    assert_optimal_termination(results)

    # TODO: incorporate costing when merged
    # add_costing(m)
    # assert_degrees_of_freedom(m, 0)
    # m.fs.costing.initialize()
    #
    # results = solve(m)
    # assert_optimal_termination(results)
    return m


def solve_flowsheet(flowsheet=None):
    """
    Solves the initial flowsheet.
    """
    fs = flowsheet
    results = solve(fs)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import watertap.property_models.NaCl_T_dep_prop_pack as props
from idaes.models.properties.tests.test_harness import (
    PropertyTestHarness as PropertyTestHarness_idaes,
)
from watertap.property_models.tests.property_test_harness import (
    PropertyTestHarness,
    PropertyRegressionTest,
)

# -----------------------------------------------------------------------------
@pytest.mark.unit
class TestNaClProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestNaClProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}
        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
        }
        self.stateblock_statistics = {
            "number_variables": 24,
            "number_total_constraints": 20,
            "number_unused_variables": 1,
            "default_degrees_of_freedom": 3,
        }  # 4 state vars, but pressure is not active
        self.default_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.965,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.035,
            ("dens_mass_phase", "Liq"): 1021.2,
            ("flow_vol_phase", "Liq"): 9.790e-4,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 985.5,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 35.74,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 53.57,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 0.5989,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9889,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 1.106e-2,
            ("molality_phase_comp", ("Liq", "NaCl")): 0.6206,
            ("osm_coeff", None): 0.9294,
            ("pressure_osm_phase", "Liq"): 2.851e6,
            ("diffus_phase_comp", ("Liq", "NaCl")): 1.517e-9,
            ("visc_d_phase", "Liq"): 0.000967,
            ("enth_mass_phase", "Liq"): 101.09e3,
            ("cp_mass_phase", "Liq"): 4000,
            ("solubility_comp", "NaCl"): 0.265,
            ("therm_cond_phase", "Liq"): 0.6015,
            ("pressure_sat", None): 0.00293e6,
        }


@pytest.mark.component
class TestNaClPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.95,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.05,
            ("temperature", None): 273.15 + 50,
            ("pressure", None): 5e5,
        }
        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.95,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("dens_mass_phase", "Liq"): 1021.7,
            ("flow_vol_phase", "Liq"): 9.787e-4,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 970.6,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 51.08,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 52.73,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 0.8556,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9840,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 1.597e-2,
            ("molality_phase_comp", ("Liq", "NaCl")): 0.9006,
            ("diffus_phase_comp", ("Liq", "NaCl")): 2.3e-9,
            ("visc_d_phase", "Liq"): 0.0005802,
            ("osm_coeff", None): 0.9322,
            ("pressure_osm_phase", "Liq"): 4.4579e6,
            ("enth_mass_phase", "Liq"): 197.9e3,
            ("cp_mass_phase", "Liq"): 3941.9,
            ("solubility_comp", "NaCl"): 0.2686,
            ("therm_cond_phase", "Liq"): 0.6392,
            ("pressure_sat", None): 0.01224e6,
        }


@pytest.mark.component
class TestNaClPropertySolution_2(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.85,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.15,
            ("temperature", None): 273.15 + 100,
            ("pressure", None): 5e5,
        }
        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.85,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.15,
            ("dens_mass_phase", "Liq"): 1063.3,
            ("flow_vol_phase", "Liq"): 0.0009404,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 903.8,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 159.4,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 47.18,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 2.566,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9484,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.05159,
            ("molality_phase_comp", ("Liq", "NaCl")): 3.019,
            ("diffus_phase_comp", ("Liq", "NaCl")): 4.577e-9,
            ("visc_d_phase", "Liq"): 0.0004139,
            ("osm_coeff", None): 1.045,
            ("pressure_osm_phase", "Liq"): 1.877e7,
            ("enth_mass_phase", "Liq"): 353.9e3,
            ("cp_mass_phase", "Liq"): 3603.7,
            ("solubility_comp", "NaCl"): 0.2799,
            ("therm_cond_phase", "Liq"): 0.6648,
            ("pressure_sat", None): 0.09011e6,
        }


@pytest.mark.component
class TestNaClPropertySolution_3(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.74,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.26,
            ("temperature", None): 273.15 + 150,
            ("pressure", None): 5e5,
        }
        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.74,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.26,
            ("dens_mass_phase", "Liq"): 1113.8,
            ("flow_vol_phase", "Liq"): 0.0008977,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 824.2,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 289.6,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 41.07,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 4.449,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9022,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.09772,
            ("molality_phase_comp", ("Liq", "NaCl")): 6.0121,
            ("diffus_phase_comp", ("Liq", "NaCl")): 4.8422e-9,
            ("visc_d_phase", "Liq"): 0.0003578,
            ("osm_coeff", None): 1.134,
            ("pressure_osm_phase", "Liq"): 4.399e7,
            ("enth_mass_phase", "Liq"): 475.3e3,
            ("cp_mass_phase", "Liq"): 3422.6,
            ("solubility_comp", "NaCl"): 0.2966,
            ("therm_cond_phase", "Liq"): 0.6627,
            ("pressure_sat", None): 0.3688e6,
        }



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

""""
Base constraints and methods for membrane distillation channel
__author__ = "Elmira Shamlou"

"""
from copy import deepcopy
from enum import Enum, auto
from pyomo.environ import (
    Constraint,
    Param,
    NonNegativeReals,
    Var,
    units as pyunits,
)
from idaes.core import (
    FlowDirection,
)
from idaes.core.util import scaling as iscale
from idaes.core.util.exceptions import ConfigurationError


class TemperaturePolarizationType(Enum):
    """
    none: no temperature polarization
    fixed: convective heat transfer coefficient is a user specified value
    calculated: calculate convective heat transfer coefficient
    """

    none = auto()
    fixed = auto()
    calculated = auto()


class ConcentrationPolarizationType(Enum):
    """
    none: no concentration polarization
    fixed: concentration polarization modulus is a user specified value
    calculated: calculate concentration polarization (concentration at membrane interface)
    """

    none = auto()
    fixed = auto()
    calculated = auto()


class MassTransferCoefficient(Enum):
    """
    none: mass transfer coefficient not utilized for concentration polarization effect
    fixed: mass transfer coefficient is a user specified value
    calculated: mass transfer coefficient is calculated
    """

    none = auto()
    fixed = auto()
    calculated = auto()
    # TODO: add option for users to define their own relationship?


class PressureChangeType(Enum):
    """
    fixed_per_stage: pressure drop across membrane channel is a user-specified value
    fixed_per_unit_length: pressure drop per unit length across membrane channel is a user-specified value
    calculated: pressure drop across membrane channel is calculated
    """

    fixed_per_stage = auto()
    fixed_per_unit_length = auto()
    calculated = auto()


class FrictionFactor(Enum):
    """
    flat_sheet: Darcy's friction factor correlation by Guillen & Hoek
    spiral_wound: Darcy's friction factor correlation by Schock & Miquel, 1987
    """

    flat_sheet = auto()
    spiral_wound = auto()


class MDChannelMixin:
    def _skip_element(self, x):
        raise NotImplementedError()

    def _set_nfe(self):
        self.first_element = self.length_domain.first()
        self.last_element = self.length_domain.last()

        self.nfe = Param(
            initialize=(len(self.difference_elements)),
            units=pyunits.dimensionless,
            doc="Number of finite elements",
        )

    def add_total_pressure_balances(
        self,
        has_pressure_change=True,
        pressure_change_type=PressureChangeType.calculated,
        custom_term=None,
        friction_factor=FrictionFactor.flat_sheet,
    ):
        super().add_total_pressure_balances(
            has_pressure_change=has_pressure_change, custom_term=custom_term
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Pressure at interface",
        )
        def eq_equal_pressure_interface(b, t, x):
            if b._skip_element(x):
                return Constraint.Skip
            return b.properties_interface[t, x].pressure == b.properties[t, x].pressure

        if has_pressure_change:
            self._add_pressure_change(pressure_change_type=pressure_change_type)

        if pressure_change_type == PressureChangeType.calculated:
            self._add_calculated_pressure_change(friction_factor=friction_factor)

    def add_temperature_polarization(
        self,
        temperature_polarization_type=TemperaturePolarizationType.none,
    ):

        units_meta = self.config.property_package.get_metadata().get_derived_units

        if temperature_polarization_type == TemperaturePolarizationType.none:

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="No temperature polarization",
            )
            def eq_no_temp_pol(b, t, x):
                if b._skip_element(x):
                    return Constraint.Skip
                return (
                    b.properties_interface[t, x].temperature
                    == b.properties[t, x].temperature
                )

            return self.eq_no_temp_pol

        elif temperature_polarization_type not in (
            TemperaturePolarizationType.fixed,
            TemperaturePolarizationType.calculated,
        ):
            raise ConfigurationError(
                f"Unrecognized temperature_polarization_type {temperature_polarization_type}"
            )

        self.h_conv = Var(
            self.flowsheet().config.time,
            self.length_domain,
            initialize=2000,
            bounds=(1, 5e4),
            domain=NonNegativeReals,
            units=pyunits.J * pyunits.s**-1 * pyunits.K**-1 * pyunits.m**-2,
            doc="Convective heat transfer coefficient",
        )

        if temperature_polarization_type == TemperaturePolarizationType.calculated:
            self._add_calculated_convective_heat_transfer_coefficient()

        return self.h_conv

    def _add_calculated_convective_heat_transfer_coefficient(self):
        self._add_calculated_pressure_change_mass_heat_transfer_components()

        self.N_Pr = Var(
            self.flowsheet().config.time,
            self.length_domain,
            initialize=1,
            bounds=(0.1, 50),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Prandtl number in membrane channel",
        )
        self.N_Nu = Var(
            self.flowsheet().config.time,
            self.length_domain,
            initialize=1e2,
            bounds=(1, 3e2),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Nusselt number in membrane channel",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Convective heat transfer coefficient",
        )
        def eq_h_conv(b, t, x):
            if b._skip_element(x):
                return Constraint.Skip
            return (
                b.h_conv[t, x]
                == b.properties[t, x].therm_cond_phase["Liq"] * b.N_Nu[t, x] / b.dh
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Nusselt number",
        )
        def eq_N_Nu(b, t, x):
            if b._skip_element(x):
                return Constraint.Skip
            return b.N_Nu[t, x] == 0.162 * b.N_Re[t, x] ** 0.656 * b.N_Pr[t, x] ** 0.333

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Prandtl number",
        )
        def eq_N_Pr(b, t, x):
            if b._skip_element(x):
                return Constraint.Skip
            return (
                b.N_Pr[t, x] * b.properties[t, x].therm_cond_phase["Liq"]
                == b.properties[t, x].visc_d_phase["Liq"]
                * b.properties[t, x].cp_mass_phase["Liq"]
            )

        return self.eq_h_conv

    def add_concentration_polarization(
        self,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
    ):

        solute_set = self.config.property_package.solute_set
        units_meta = self.config.property_package.get_metadata().get_derived_units

        if concentration_polarization_type == ConcentrationPolarizationType.none:

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                solute_set,
                doc="Unit concentration polarization modulus",
            )
            def eq_cp_modulus(b, t, x, j):
                if b._skip_element(x):
                    return Constraint.Skip
                return (
                    b.properties_interface[t, x].conc_mass_phase_comp["Liq", j]
                    == b.properties[t, x].conc_mass_phase_comp["Liq", j]
                )

            return self.eq_cp_modulus

        if concentration_polarization_type not in (
            ConcentrationPolarizationType.fixed,
            ConcentrationPolarizationType.calculated,
        ):
            raise ConfigurationError(
                f"Unrecognized concentration_polarization_type {concentration_polarization_type}"
            )

        self.cp_modulus = Var(
            self.flowsheet().config.time,
            self.length_domain,
            solute_set,
            initialize=1.1,
            bounds=(0.1, 3),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Concentration polarization modulus",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            solute_set,
            doc="Concentration polarization modulus",
        )
        def eq_cp_modulus(b, t, x, j):  # pylint: disable=function-redefined
            if b._skip_element(x):
                return Constraint.Skip
            return (
                self.properties_interface[t, x].conc_mass_phase_comp["Liq", j]
                == self.properties[t, x].conc_mass_phase_comp["Liq", j]
                * self.cp_modulus[t, x, j]
            )

        if concentration_polarization_type == ConcentrationPolarizationType.calculated:
            if mass_transfer_coefficient == MassTransferCoefficient.none:
                raise ConfigurationError()

            # mass_transfer_coefficient is either calculated or fixed
            self.K = Var(
                self.flowsheet().config.time,
                self.length_domain,
                solute_set,
                initialize=5e-5,
                bounds=(1e-6, 1e-3),
                domain=NonNegativeReals,
                units=units_meta("length") * units_meta("time") ** -1,
                doc="Membrane channel mass transfer coefficient",
            )

            if mass_transfer_coefficient == MassTransferCoefficient.calculated:
                self._add_calculated_mass_transfer_coefficient()

        return self.eq_cp_modulus

    ## should be called by add concentration polarization
    def _add_calculated_mass_transfer_coefficient(self):
        self._add_calculated_pressure_change_mass_heat_transfer_components()

        solute_set = self.config.property_package.solute_set

        self.N_Sc_comp = Var(
            self.flowsheet().config.time,
            self.length_domain,
            solute_set,
            initialize=5e2,
            bounds=(1e2, 2e3),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Schmidt number in membrane channel",
        )
        self.N_Sh_comp = Var(
            self.flowsheet().config.time,
            self.length_domain,
            solute_set,
            initialize=1e2,
            bounds=(1, 3e2),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Sherwood number in membrane channel",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            self.config.property_package.solute_set,
            doc="Mass transfer coefficient in membrane channel",
        )
        def eq_K(b, t, x, j):
            if b._skip_element(x):
                return Constraint.Skip
            return (
                b.K[t, x, j] * b.dh
                # TODO: add diff coefficient to SW prop and consider multi-components
                == b.properties[t, x].diffus_phase_comp["Liq", j] * b.N_Sh_comp[t, x, j]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            self.config.property_package.solute_set,
            doc="Sherwood number",
        )
        def eq_N_Sh_comp(b, t, x, j):
            return (
                b.N_Sh_comp[t, x, j]
                == 0.2 * b.N_Re[t, x] ** 0.57 * b.N_Sc_comp[t, x, j] ** 0.4
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            self.config.property_package.solute_set,
            doc="Schmidt number",
        )
        def eq_N_Sc_comp(b, t, x, j):
            return (
                b.N_Sc_comp[t, x, j]
                * b.properties[t, x].dens_mass_phase["Liq"]
                * b.properties[t, x].diffus_phase_comp["Liq", j]
                == b.properties[t, x].visc_d_phase["Liq"]
            )

        return self.eq_K

    def _add_calculated_pressure_change_mass_heat_transfer_components(self):
        # NOTE: This function could be called by either
        # `_add_calculated_pressure_change` *and/or*
        # `_add_calculated_mass_transfer_coefficient`.
        # Therefore, we add this simple guard against it being called twice.
        if hasattr(self, "channel_height"):
            return

        if not hasattr(self, "width"):
            raise ConfigurationError(
                f"Due to either a calculated mass transfer coefficient or a calculated pressure change, a ``width`` variable needs to be supplied to `add_geometry` for this MembraneChannel"
            )

        units_meta = self.config.property_package.get_metadata().get_derived_units

        if not hasattr(self, "area"):
            # comes from ControlVolume1D for 1DMC
            self.area = Var(
                initialize=1e-3 * 1 * 0.95,
                bounds=(0, 1e3),
                domain=NonNegativeReals,
                units=units_meta("length") ** 2,
                doc="Cross sectional area",
            )

        self.channel_height = Var(
            initialize=1e-3,
            bounds=(1e-4, 5e-3),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="membrane-channel height",
        )

        self.dh = Var(
            initialize=1e-3,
            bounds=(1e-4, 5e-3),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="Hydraulic diameter of membrane channel",
        )

        self.spacer_porosity = Var(
            initialize=0.95,
            bounds=(0.1, 0.99),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="membrane-channel spacer porosity",
        )

        self.N_Re = Var(
            self.flowsheet().config.time,
            self.length_domain,
            initialize=5e2,
            bounds=(10, 5e3),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Reynolds number in membrane channel",
        )

        @self.Constraint(doc="Hydraulic diameter")  # eqn. 17 in Schock & Miquel, 1987
        def eq_dh(b):
            return b.dh == 4 * b.spacer_porosity / (
                2 / b.channel_height + (1 - b.spacer_porosity) * 8 / b.channel_height
            )

        @self.Constraint(doc="Cross-sectional area")
        def eq_area(b):
            return b.area == b.channel_height * b.width * b.spacer_porosity

        @self.Constraint(
            self.flowsheet().config.time, self.length_domain, doc="Reynolds number"
        )
        def eq_N_Re(b, t, x):
            return (
                b.N_Re[t, x] * b.area * b.properties[t, x].visc_d_phase["Liq"]
                == sum(
                    b.properties[t, x].flow_mass_phase_comp["Liq", j]
                    for j in b.config.property_package.component_list
                )
                * b.dh
            )

    def _add_calculated_pressure_change(
        self, friction_factor=FrictionFactor.flat_sheet
    ):
        self._add_calculated_pressure_change_mass_heat_transfer_components()

        units_meta = self.config.property_package.get_metadata().get_derived_units

        self.velocity = Var(
            self.flowsheet().config.time,
            self.length_domain,
            initialize=0.5,
            bounds=(1e-2, 5),
            domain=NonNegativeReals,
            units=units_meta("length") / units_meta("time"),
            doc="Crossflow velocity in feed channel",
        )
        self.friction_factor_darcy = Var(
            self.flowsheet().config.time,
            self.length_domain,
            initialize=0.5,
            bounds=(1e-2, 5),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Darcy friction factor in feed channel",
        )

        # Constraints active when MassTransferCoefficient.calculated
        # Mass transfer coefficient calculation

        ## ==========================================================================
        # Crossflow velocity
        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Crossflow velocity constraint",
        )
        def eq_velocity(b, t, x):
            return b.velocity[t, x] * b.area == b.properties[t, x].flow_vol_phase["Liq"]

        ## ==========================================================================
        # Darcy friction factor based on eq. S27 in SI for Cost Optimization of Osmotically Assisted Reverse Osmosis
        if friction_factor == FrictionFactor.flat_sheet:

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Darcy friction factor constraint for flat sheet membranes",
            )
            def eq_friction_factor(b, t, x):
                return (b.friction_factor_darcy[t, x] - 0.42) * b.N_Re[t, x] == 189.3

        # Darcy friction factor based on eq. 24 in Mass transfer and pressure loss in spiral wound modules (Schock & Miquel, 1987)
        elif friction_factor == FrictionFactor.spiral_wound:

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Darcy friction factor constraint for spiral-wound membranes",
            )
            def eq_friction_factor(b, t, x):
                return b.friction_factor_darcy[t, x] == 6.23 * b.N_Re[t, x] ** -0.3

        else:
            raise ConfigurationError(
                f"Unrecognized friction_factor type {friction_factor}"
            )

        ## ==========================================================================
        # Pressure change per unit length due to friction
        # -1/2*f/dh*density*velocity^2
        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="pressure change per unit length due to friction",
        )
        def eq_dP_dx(b, t, x):
            return (
                b.dP_dx[t, x] * b.dh
                == -0.5
                * b.friction_factor_darcy[t, x]
                * b.properties[t, x].dens_mass_phase["Liq"]
                * b.velocity[t, x] ** 2
            )

    def _add_interface_stateblock(self, has_phase_equilibrium=None):
        """
        This method constructs the interface state blocks for the
        control volume.

        Args:
            has_phase_equilibrium: indicates whether equilibrium calculations
                                    will be required in state blocks
        Returns:
            None
        """
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = has_phase_equilibrium
        tmp_dict["defined_state"] = False  # these blocks are not inlets or outlets

        self.properties_interface = self.config.property_package.build_state_block(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Material properties of feed-side membrane interface",
            **tmp_dict,
        )

    def _add_vapor_stateblock(
        self,
        property_package_vapor=None,
        property_package_args_vapor=None,
        has_phase_equilibrium=None,
    ):
        """
        This method constructs the vapor state blocks for the
        control volume.
        """
        if property_package_vapor is None or property_package_args_vapor is None:
            raise ValueError(
                "Vapor property package and its arguments must be provided."
            )

        tmp_dict = dict(**property_package_args_vapor)
        tmp_dict["has_phase_equilibrium"] = has_phase_equilibrium
        tmp_dict["defined_state"] = False  # these blocks are not inlets or outlets

        self.properties_vapor = property_package_vapor.build_state_block(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Material properties of vapor phase",
            **tmp_dict,
        )

    def add_extensive_flow_to_interface(self):
        # VOLUMETRIC FLOWRATE
        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Volumetric flow at interface of inlet",
        )
        def eq_equal_flow_vol_interface(b, t, x):
            if b._skip_element(x):
                return Constraint.Skip
            return (
                b.properties_interface[t, x].flow_vol_phase["Liq"]
                == b.properties[t, x].flow_vol_phase["Liq"]
            )

    def _get_state_args(self, initialize_guess, state_args):
        """
        Arguments:
            initialize_guess : a dict of guesses for recovery, Temp_change, and deltaP.
            state_args : a dict of arguments to be passed to the property
                        package to provide an initial state for the inlet feed.
        """

        if initialize_guess is None:
            initialize_guess = {}

        if "recovery" not in initialize_guess:
            initialize_guess["recovery"] = 0.1

        if "deltaP" not in initialize_guess:
            initialize_guess["deltaP"] = -5e5

        Temp_change = initialize_guess.get("Temp_change", 40)

        # Getting source state
        if self.flow_direction == FlowDirection.forward:
            source_idx = self.length_domain.first()
        else:
            source_idx = self.length_domain.last()
        source = self.properties[self.flowsheet().config.time.first(), source_idx]

        if state_args is None:
            state_args = {}
            state_dict = source.define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        hot_outlet_args = deepcopy(state_args)
        cold_outlet_args = deepcopy(state_args)

        # Adjust flow based on recovery
        for j in self.config.property_package.component_list:
            hot_outlet_args["flow_mass_phase_comp"]["Liq", j] *= (
                1 - initialize_guess["recovery"]
            )
            cold_outlet_args["flow_mass_phase_comp"]["Liq", j] *= (
                1 + initialize_guess["recovery"]
            )

        # Adjust temperature based on Temp_change_ratio
        hot_outlet_args["temperature"] = state_args["temperature"] - Temp_change
        cold_outlet_args["temperature"] = state_args["temperature"] + Temp_change

        return {
            "inlet": state_args,
            "hot_outlet": hot_outlet_args,
            "cold_outlet": cold_outlet_args,
        }

    def _get_average_state(self, prop_in, prop_out):
        """Helper function to compute average of two property states"""
        x = 0.5
        state_args_tx = {}

        for k in prop_in:
            if isinstance(prop_in[k], dict):
                if k not in state_args_tx:
                    state_args_tx[k] = {}
                for index in prop_in[k]:
                    state_args_tx[k][index] = (1.0 - x) * prop_in[k][
                        index
                    ] + x * prop_out[k][index]
            else:
                state_args_tx[k] = (1.0 - x) * prop_in[k] + x * prop_out[k]

        return state_args_tx

    def _get_state_args_interface(self, prop_in, prop_out):
        return self._get_average_state(prop_in, prop_out)

    def _get_state_args_vapor(self, prop_in, prop_out):
        state_args = self._get_average_state(prop_in, prop_out)

        state_args_vapor = {}

        # state_args_vapor["pressure"] = state_args["pressure"]
        state_args_vapor["temperature"] = state_args["temperature"]

        state_args_vapor["flow_mass_phase_comp"] = {
            ("Liq", "H2O"): self.properties_vapor[0, 0]
            .flow_mass_phase_comp["Liq", "H2O"]
            .lb,
            ("Vap", "H2O"): state_args["flow_mass_phase_comp"][("Liq", "H2O")] / 20,
        }

        return state_args_vapor

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if hasattr(self, "channel_height"):
            if iscale.get_scaling_factor(self.channel_height) is None:
                iscale.set_scaling_factor(self.channel_height, 1e3)

        if hasattr(self, "spacer_porosity"):
            if iscale.get_scaling_factor(self.spacer_porosity) is None:
                iscale.set_scaling_factor(self.spacer_porosity, 1)

        if hasattr(self, "dh"):
            if iscale.get_scaling_factor(self.dh) is None:
                iscale.set_scaling_factor(self.dh, 1e3)

        if hasattr(self, "K"):
            for v in self.K.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e4)

        if hasattr(self, "N_Re"):
            for t, x in self.N_Re.keys():
                if iscale.get_scaling_factor(self.N_Re[t, x]) is None:
                    iscale.set_scaling_factor(self.N_Re[t, x], 1e-2)

        if hasattr(self, "N_Sc_comp"):
            for v in self.N_Sc_comp.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-2)

        if hasattr(self, "N_Sh_comp"):
            for v in self.N_Sh_comp.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-2)

        if hasattr(self, "N_Pr"):
            for v in self.N_Pr.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, "N_Nu"):
            for v in self.N_Nu.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, "h_conv"):
            for (t, x), v in self.h_conv.items():
                if iscale.get_scaling_factor(v) is None:
                    if hasattr(self, "N_Nu"):
                        sf_hconv = (
                            iscale.get_scaling_factor(
                                self.properties[t, x].therm_cond_phase["Liq"]
                            )
                            * iscale.get_scaling_factor(self.N_Nu[t, x])
                            / iscale.get_scaling_factor(self.dh)
                        )
                        iscale.set_scaling_factor(v, sf_hconv)
                    else:
                        iscale.set_scaling_factor(v, 1e3)

        if hasattr(self, "velocity"):
            for v in self.velocity.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, "friction_factor_darcy"):
            for v in self.friction_factor_darcy.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, "cp_modulus"):
            for v in self.cp_modulus.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)
